@@ -137,7 +137,7 @@ Failed to parse riotfile 'riotfile.py'.
     assert (
         """
 SyntaxError: invalid syntax
-""".lstrip()
+""".strip()
         in result.stderr
     )
     assert result.returncode == 1
@@ -433,6 +433,7 @@ def test_failure():
 """,
     )
     result = tmp_run("riot run -s pass")
+    assert result.returncode == 0, result.stderr
     assert re.search(
         r"""
 ============================= test session starts ==============================
@@ -448,9 +449,8 @@ test_success.py .*
 ✓ pass: .*
 1 passed with 0 warnings, 0 failed\n""".lstrip(),
         result.stdout,
-    )
+    ), result.stdout
     assert result.stderr == ""
-    assert result.returncode == 0
 
     result = tmp_run("riot run -s fail")
     assert (
@@ -702,3 +702,50 @@ venv = Venv(
     assert env["foo"] == "40"
     assert env["bar"] == "43"
     assert env["baz"] == "stillthesame"
+
+
+def test_create(tmp_path: pathlib.Path, tmp_run: _T_TmpRun) -> None:
+    rf_path = tmp_path / "riotfile.py"
+    rf_path.write_text(
+        """
+from riot import Venv
+venv = Venv(
+    name="test",
+    pys=[3],
+    pkgs={"pip": ""},
+    venvs=[
+        Venv(
+            name="child",
+            pkgs={"pytest": ""},
+            create=True,
+            command="echo $PYTHONPATH",
+        ),
+    ],
+)
+""",
+    )
+    result = tmp_run("riot -v run -s child")
+    venv_path = tmp_path / ".riot/venv_py{}_pip_pytest".format(
+        "".join((str(_) for _ in sys.version_info[:3]))
+    )
+    assert f"Creating virtualenv '{venv_path}'" in result.stderr
+    assert f"Running command 'echo $PYTHONPATH' in venv '{venv_path}'"
+    assert result.stdout.startswith(":".join(("", str(tmp_path))))
+    assert result.returncode == 0
+
+
+def test_extras(tmp_path: pathlib.Path, tmp_run: _T_TmpRun) -> None:
+    rf_path = tmp_path / "riotfile.py"
+    rf_path.write_text(
+        """
+from riot import Venv, latest
+venv = Venv(
+    name="test",
+    pys=["3"],
+    pkgs={"requests[socks]": latest},
+    command="python -c 'import requests'",
+)
+""",
+    )
+    result = tmp_run("riot -v run -s test")
+    assert result.returncode == 0
